// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// ServiceLoadBalancerIPBlockApplyConfiguration represents a declarative configuration of the ServiceLoadBalancerIPBlock type for use
// with apply.
//
// ServiceLoadBalancerIPBlock represents a single allowed LoadBalancer IP CIDR block.
type ServiceLoadBalancerIPBlockApplyConfiguration struct {
	CIDR *string `json:"cidr,omitempty"`
}

// ServiceLoadBalancerIPBlockApplyConfiguration constructs a declarative configuration of the ServiceLoadBalancerIPBlock type for use with
// apply.
func ServiceLoadBalancerIPBlock() *ServiceLoadBalancerIPBlockApplyConfiguration {
	return &ServiceLoadBalancerIPBlockApplyConfiguration{}
}

// WithCIDR sets the CIDR field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the CIDR field is set to the value of the last call.
func (b *ServiceLoadBalancerIPBlockApplyConfiguration) WithCIDR(value string) *ServiceLoadBalancerIPBlockApplyConfiguration {
	b.CIDR = &value
	return b
}
