// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// CalicoNodeAgentStatusApplyConfiguration represents a declarative configuration of the CalicoNodeAgentStatus type for use
// with apply.
//
// CalicoNodeAgentStatus defines the observed state of agent status on the node.
type CalicoNodeAgentStatusApplyConfiguration struct {
	// BIRDV4 represents the latest observed status of bird4.
	BIRDV4 *BGPDaemonStatusApplyConfiguration `json:"birdV4,omitempty"`
	// BIRDV6 represents the latest observed status of bird6.
	BIRDV6 *BGPDaemonStatusApplyConfiguration `json:"birdV6,omitempty"`
}

// CalicoNodeAgentStatusApplyConfiguration constructs a declarative configuration of the CalicoNodeAgentStatus type for use with
// apply.
func CalicoNodeAgentStatus() *CalicoNodeAgentStatusApplyConfiguration {
	return &CalicoNodeAgentStatusApplyConfiguration{}
}

// WithBIRDV4 sets the BIRDV4 field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the BIRDV4 field is set to the value of the last call.
func (b *CalicoNodeAgentStatusApplyConfiguration) WithBIRDV4(value *BGPDaemonStatusApplyConfiguration) *CalicoNodeAgentStatusApplyConfiguration {
	b.BIRDV4 = value
	return b
}

// WithBIRDV6 sets the BIRDV6 field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the BIRDV6 field is set to the value of the last call.
func (b *CalicoNodeAgentStatusApplyConfiguration) WithBIRDV6(value *BGPDaemonStatusApplyConfiguration) *CalicoNodeAgentStatusApplyConfiguration {
	b.BIRDV6 = value
	return b
}
