// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by client-gen. DO NOT EDIT.

package fake

import (
	v3 "github.com/projectcalico/api/pkg/client/clientset_generated/clientset/typed/projectcalico/v3"
	rest "k8s.io/client-go/rest"
	testing "k8s.io/client-go/testing"
)

type FakeProjectcalicoV3 struct {
	*testing.Fake
}

func (c *FakeProjectcalicoV3) BGPConfigurations() v3.BGPConfigurationInterface {
	return newFakeBGPConfigurations(c)
}

func (c *FakeProjectcalicoV3) BGPFilters() v3.BGPFilterInterface {
	return newFakeBGPFilters(c)
}

func (c *FakeProjectcalicoV3) BGPPeers() v3.BGPPeerInterface {
	return newFakeBGPPeers(c)
}

func (c *FakeProjectcalicoV3) BlockAffinities() v3.BlockAffinityInterface {
	return newFakeBlockAffinities(c)
}

func (c *FakeProjectcalicoV3) CalicoNodeStatuses() v3.CalicoNodeStatusInterface {
	return newFakeCalicoNodeStatuses(c)
}

func (c *FakeProjectcalicoV3) ClusterInformations() v3.ClusterInformationInterface {
	return newFakeClusterInformations(c)
}

func (c *FakeProjectcalicoV3) FelixConfigurations() v3.FelixConfigurationInterface {
	return newFakeFelixConfigurations(c)
}

func (c *FakeProjectcalicoV3) GlobalNetworkPolicies() v3.GlobalNetworkPolicyInterface {
	return newFakeGlobalNetworkPolicies(c)
}

func (c *FakeProjectcalicoV3) GlobalNetworkSets() v3.GlobalNetworkSetInterface {
	return newFakeGlobalNetworkSets(c)
}

func (c *FakeProjectcalicoV3) HostEndpoints() v3.HostEndpointInterface {
	return newFakeHostEndpoints(c)
}

func (c *FakeProjectcalicoV3) IPAMBlocks() v3.IPAMBlockInterface {
	return newFakeIPAMBlocks(c)
}

func (c *FakeProjectcalicoV3) IPAMConfigurations() v3.IPAMConfigurationInterface {
	return newFakeIPAMConfigurations(c)
}

func (c *FakeProjectcalicoV3) IPAMHandles(namespace string) v3.IPAMHandleInterface {
	return newFakeIPAMHandles(c, namespace)
}

func (c *FakeProjectcalicoV3) IPPools() v3.IPPoolInterface {
	return newFakeIPPools(c)
}

func (c *FakeProjectcalicoV3) IPReservations() v3.IPReservationInterface {
	return newFakeIPReservations(c)
}

func (c *FakeProjectcalicoV3) KubeControllersConfigurations() v3.KubeControllersConfigurationInterface {
	return newFakeKubeControllersConfigurations(c)
}

func (c *FakeProjectcalicoV3) NetworkPolicies(namespace string) v3.NetworkPolicyInterface {
	return newFakeNetworkPolicies(c, namespace)
}

func (c *FakeProjectcalicoV3) NetworkSets(namespace string) v3.NetworkSetInterface {
	return newFakeNetworkSets(c, namespace)
}

func (c *FakeProjectcalicoV3) Profiles() v3.ProfileInterface {
	return newFakeProfiles(c)
}

func (c *FakeProjectcalicoV3) StagedGlobalNetworkPolicies() v3.StagedGlobalNetworkPolicyInterface {
	return newFakeStagedGlobalNetworkPolicies(c)
}

func (c *FakeProjectcalicoV3) StagedKubernetesNetworkPolicies(namespace string) v3.StagedKubernetesNetworkPolicyInterface {
	return newFakeStagedKubernetesNetworkPolicies(c, namespace)
}

func (c *FakeProjectcalicoV3) StagedNetworkPolicies(namespace string) v3.StagedNetworkPolicyInterface {
	return newFakeStagedNetworkPolicies(c, namespace)
}

func (c *FakeProjectcalicoV3) Tiers() v3.TierInterface {
	return newFakeTiers(c)
}

// RESTClient returns a RESTClient that is used to communicate
// with API server by this client implementation.
func (c *FakeProjectcalicoV3) RESTClient() rest.Interface {
	var ret *rest.RESTClient
	return ret
}
