// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// NetworkSetSpecApplyConfiguration represents a declarative configuration of the NetworkSetSpec type for use
// with apply.
//
// NetworkSetSpec contains the specification for a NetworkSet resource.
type NetworkSetSpecApplyConfiguration struct {
	// The list of IP networks that belong to this set.
	Nets []string `json:"nets,omitempty"`
}

// NetworkSetSpecApplyConfiguration constructs a declarative configuration of the NetworkSetSpec type for use with
// apply.
func NetworkSetSpec() *NetworkSetSpecApplyConfiguration {
	return &NetworkSetSpecApplyConfiguration{}
}

// WithNets adds the given value to the Nets field in the declarative configuration
// and returns the receiver, so that objects can be build by chaining "With" function invocations.
// If called multiple times, values provided by each call will be appended to the Nets field.
func (b *NetworkSetSpecApplyConfiguration) WithNets(values ...string) *NetworkSetSpecApplyConfiguration {
	for i := range values {
		b.Nets = append(b.Nets, values[i])
	}
	return b
}
