export const tabTitleTextStyles = (hasNodata: boolean) => ({
    fontSize: 'xs',
    fontWeight: 600,
    color: hasNodata ? 'tigeraGrey.500' : 'inherit',
});

export const tagStyles = {
    paddingInlineEnd: 2,
    paddingInlineStart: 2,
    minWidth: 'unset',
    alignItems: 'center',
    ml: 2,
    fontSize: 'xs',
};
