// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// BGPFilterSetPriorityApplyConfiguration represents a declarative configuration of the BGPFilterSetPriority type for use
// with apply.
//
// BGPFilterSetPriority specifies a route priority to set.
type BGPFilterSetPriorityApplyConfiguration struct {
	// Value is the priority to set, in the same units as FelixConfiguration's
	// ...RoutePriority fields.
	Value *int `json:"value,omitempty"`
}

// BGPFilterSetPriorityApplyConfiguration constructs a declarative configuration of the BGPFilterSetPriority type for use with
// apply.
func BGPFilterSetPriority() *BGPFilterSetPriorityApplyConfiguration {
	return &BGPFilterSetPriorityApplyConfiguration{}
}

// WithValue sets the Value field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Value field is set to the value of the last call.
func (b *BGPFilterSetPriorityApplyConfiguration) WithValue(value int) *BGPFilterSetPriorityApplyConfiguration {
	b.Value = &value
	return b
}
