//go:build !ignore_autogenerated
// +build !ignore_autogenerated

// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by openapi-gen. DO NOT EDIT.

package numorstring

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in DSCP) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.lib.numorstring.DSCP"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Port) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.lib.numorstring.Port"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Protocol) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.lib.numorstring.Protocol"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Uint8OrString) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.lib.numorstring.Uint8OrString"
}
