// Copyright (c) 2024-2025 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package dataplanedefs

import "github.com/vishvananda/netlink"

const (
	IPIPIfaceName    = "tunl0"
	VXLANIfaceNameV4 = "vxlan.calico"
	VXLANIfaceNameV6 = "vxlan-v6.calico"

	DefaultRouteProto netlink.RouteProtocol = 80

	BPFInDev  = "bpfin.cali"
	BPFOutDev = "bpfout.cali"
)
