// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// RouteTableRangeApplyConfiguration represents a declarative configuration of the RouteTableRange type for use
// with apply.
type RouteTableRangeApplyConfiguration struct {
	Min *int `json:"min,omitempty"`
	Max *int `json:"max,omitempty"`
}

// RouteTableRangeApplyConfiguration constructs a declarative configuration of the RouteTableRange type for use with
// apply.
func RouteTableRange() *RouteTableRangeApplyConfiguration {
	return &RouteTableRangeApplyConfiguration{}
}

// WithMin sets the Min field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Min field is set to the value of the last call.
func (b *RouteTableRangeApplyConfiguration) WithMin(value int) *RouteTableRangeApplyConfiguration {
	b.Min = &value
	return b
}

// WithMax sets the Max field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Max field is set to the value of the last call.
func (b *RouteTableRangeApplyConfiguration) WithMax(value int) *RouteTableRangeApplyConfiguration {
	b.Max = &value
	return b
}
