// Code generated by mockery v2.52.3. DO NOT EDIT.

package mocks

import (
	context "context"

	mock "github.com/stretchr/testify/mock"

	tunnel "github.com/projectcalico/calico/guardian/pkg/tunnel"
)

// SessionDialer is an autogenerated mock type for the SessionDialer type
type SessionDialer struct {
	mock.Mock
}

// Dial provides a mock function with given fields: ctx
func (_m *SessionDialer) Dial(ctx context.Context) (tunnel.Session, error) {
	ret := _m.Called(ctx)

	if len(ret) == 0 {
		panic("no return value specified for Dial")
	}

	var r0 tunnel.Session
	var r1 error
	if rf, ok := ret.Get(0).(func(context.Context) (tunnel.Session, error)); ok {
		return rf(ctx)
	}
	if rf, ok := ret.Get(0).(func(context.Context) tunnel.Session); ok {
		r0 = rf(ctx)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(tunnel.Session)
		}
	}

	if rf, ok := ret.Get(1).(func(context.Context) error); ok {
		r1 = rf(ctx)
	} else {
		r1 = ret.Error(1)
	}

	return r0, r1
}

// NewSessionDialer creates a new instance of SessionDialer. It also registers a testing interface on the mock and a cleanup function to assert the mocks expectations.
// The first argument is typically a *testing.T value.
func NewSessionDialer(t interface {
	mock.TestingT
	Cleanup(func())
}) *SessionDialer {
	mock := &SessionDialer{}
	mock.Mock.Test(t)

	t.Cleanup(func() { mock.AssertExpectations(t) })

	return mock
}
