export const checkboxStyles = {
    cursor: 'pointer',
    border: '1px solid',
    borderColor: 'tigeraGrey.600',
    color: 'tigeraGrey.400',
    py: 1,
    textAlign: 'center',
    transition: 'all 0.15s ease-in-out',
    alignItems: 'center',
    justifyContent: 'center',
    fontSize: 'sm',
    _hover: {
        backgroundColor: 'tigeraGrey.600',
    },
};

export const firstStyles = {
    borderLeftRadius: 'md',
    borderLeft: '1px solid',
    borderLeftColor: 'tigeraGrey.600',
    borderRightWidth: '0',
};

export const lastStyles = {
    borderRightRadius: 'md',
    borderRight: '1px solid',
    borderRightColor: 'tigeraGrey.600',
};
