export default {
    'experimental-token-category-1': {
        _light: 'experimental-color-green.600',
        _dark: 'experimental-color-green.400',
    },
    'experimental-token-category-2': {
        _light: 'experimental-color-medium-blue.500',
        _dark: 'experimental-color-medium-blue.400',
    },
    'experimental-token-category-3': {
        _light: 'experimental-color-medium-gold.600',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-category-4': {
        _light: 'experimental-color-lilac.600',
        _dark: 'experimental-color-lilac.400',
    },
    'experimental-token-category-5': {
        _light: 'experimental-color-peach.600',
        _dark: 'experimental-color-peach.400',
    },
    'experimental-token-category-6': {
        _light: 'experimental-color-rose.600',
        _dark: 'experimental-color-rose.400',
    },
    'experimental-token-category-7': {
        _light: 'experimental-color-sky.600',
        _dark: 'experimental-color-sky.400',
    },
    'experimental-token-category-8': {
        _light: 'experimental-color-zest.600',
        _dark: 'experimental-color-zest.400',
    },
    'experimental-token-category-9': {
        _light: 'experimental-color-dark-gold.600',
        _dark: 'experimental-color-dark-gold.400',
    },
    'experimental-token-category-10': {
        _light: 'experimental-color-coast.600',
        _dark: 'experimental-color-coast.400',
    },
};
