export const dividerStyles = {
    borderRight: '1px solid',
    borderColor: 'tigera-color-outline',
    height: '30px',
    pl: 2,
};

export default {
    baseStyle: {
        display: 'flex',
        flexDirection: 'row',
        flexWrap: 'wrap',
        gap: 1,
        alignItems: 'center',
    },
};
