// Code generated by mockery v2.52.3. DO NOT EDIT.

package time

import (
	time "time"

	mock "github.com/stretchr/testify/mock"
)

// Ticker is an autogenerated mock type for the Ticker type
type Ticker struct {
	mock.Mock
}

// Chan provides a mock function with no fields
func (_m *Ticker) Chan() <-chan time.Time {
	ret := _m.Called()

	if len(ret) == 0 {
		panic("no return value specified for Chan")
	}

	var r0 <-chan time.Time
	if rf, ok := ret.Get(0).(func() <-chan time.Time); ok {
		r0 = rf()
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(<-chan time.Time)
		}
	}

	return r0
}

// Reset provides a mock function with given fields: d
func (_m *Ticker) Reset(d time.Duration) {
	_m.Called(d)
}

// Stop provides a mock function with no fields
func (_m *Ticker) Stop() {
	_m.Called()
}

// NewTicker creates a new instance of Ticker. It also registers a testing interface on the mock and a cleanup function to assert the mocks expectations.
// The first argument is typically a *testing.T value.
func NewTicker(t interface {
	mock.TestingT
	Cleanup(func())
}) *Ticker {
	mock := &Ticker{}
	mock.Mock.Test(t)

	t.Cleanup(func() { mock.AssertExpectations(t) })

	return mock
}
