export const streamButtonStyles = {
    fontSize: 'sm',
    _dark: {
        _hover: {
            bg: 'tigeraGrey.1000',
            color: 'tigera-color-primary',
            svg: {
                fill: 'tigera-color-primary',
            },
        },
        _focus: { boxShadow: 'none' },
        _active: {
            bg: 'tigeraGrey.800',
        },
    },
};

export const tabStyles = {
    fontSize: 'sm',
};
