// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

import (
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// HealthTimeoutOverrideApplyConfiguration represents a declarative configuration of the HealthTimeoutOverride type for use
// with apply.
type HealthTimeoutOverrideApplyConfiguration struct {
	Name    *string      `json:"name,omitempty"`
	Timeout *v1.Duration `json:"timeout,omitempty"`
}

// HealthTimeoutOverrideApplyConfiguration constructs a declarative configuration of the HealthTimeoutOverride type for use with
// apply.
func HealthTimeoutOverride() *HealthTimeoutOverrideApplyConfiguration {
	return &HealthTimeoutOverrideApplyConfiguration{}
}

// WithName sets the Name field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Name field is set to the value of the last call.
func (b *HealthTimeoutOverrideApplyConfiguration) WithName(value string) *HealthTimeoutOverrideApplyConfiguration {
	b.Name = &value
	return b
}

// WithTimeout sets the Timeout field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Timeout field is set to the value of the last call.
func (b *HealthTimeoutOverrideApplyConfiguration) WithTimeout(value v1.Duration) *HealthTimeoutOverrideApplyConfiguration {
	b.Timeout = &value
	return b
}
