export default {
    variants: {
        solid: {
            backgroundColor: 'experimental-token-bg-neutral',
            color: 'experimental-token-fg-default',
        },
        rounded: {
            color: 'tigeraWhite',
            backgroundColor: 'tigeraBlueDark',
            fontSize: 'xs',
            fontWeight: 'semibold',
            px: '0.375rem',
            borderRadius: '8px',
            _dark: {
                color: 'tigeraGrey.800',
                backgroundColor: 'tigeraBlueLight',
            },
        },
    },
};
