// Project Calico BPF dataplane programs.
// Copyright (c) 2021-2025 Tigera, Inc. All rights reserved.
// SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later

#ifndef __CALI_EVENTS_TYPE_H__
#define __CALI_EVENTS_TYPE_H__

#define EVENT_PROTO_STATS       1
#define EVENT_DNS               2
#define EVENT_POLICY_VERDICT    3
#define EVENT_TCP_STATS         4
#define EVENT_PROCESS_PATH      5
#define EVENT_DNS_L3            6

#define EVENT_POLICY_VERDICT_V6    7
#endif /* __CALI_EVENTS_TYPE_H__ */
