// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// BGPFilterPrefixLengthV4ApplyConfiguration represents a declarative configuration of the BGPFilterPrefixLengthV4 type for use
// with apply.
type BGPFilterPrefixLengthV4ApplyConfiguration struct {
	Min *int32 `json:"min,omitempty"`
	Max *int32 `json:"max,omitempty"`
}

// BGPFilterPrefixLengthV4ApplyConfiguration constructs a declarative configuration of the BGPFilterPrefixLengthV4 type for use with
// apply.
func BGPFilterPrefixLengthV4() *BGPFilterPrefixLengthV4ApplyConfiguration {
	return &BGPFilterPrefixLengthV4ApplyConfiguration{}
}

// WithMin sets the Min field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Min field is set to the value of the last call.
func (b *BGPFilterPrefixLengthV4ApplyConfiguration) WithMin(value int32) *BGPFilterPrefixLengthV4ApplyConfiguration {
	b.Min = &value
	return b
}

// WithMax sets the Max field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Max field is set to the value of the last call.
func (b *BGPFilterPrefixLengthV4ApplyConfiguration) WithMax(value int32) *BGPFilterPrefixLengthV4ApplyConfiguration {
	b.Max = &value
	return b
}
