export {
    listItemStyles,
    selectedOptionsHeadingStyles,
} from '../OmniCheckboxList/styles';

export const selectedIndicatorStyles = {
    width: '16px',
    height: '16px',
    flexShrink: 0,
};
