// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

import (
	projectcalicov3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
)

// BGPFilterAddCommunityApplyConfiguration represents a declarative configuration of the BGPFilterAddCommunity type for use
// with apply.
//
// BGPFilterAddCommunity specifies a BGP community to add to a route.
type BGPFilterAddCommunityApplyConfiguration struct {
	// Value is the BGP community to add.
	Value *projectcalicov3.BGPCommunityValue `json:"value,omitempty"`
}

// BGPFilterAddCommunityApplyConfiguration constructs a declarative configuration of the BGPFilterAddCommunity type for use with
// apply.
func BGPFilterAddCommunity() *BGPFilterAddCommunityApplyConfiguration {
	return &BGPFilterAddCommunityApplyConfiguration{}
}

// WithValue sets the Value field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the Value field is set to the value of the last call.
func (b *BGPFilterAddCommunityApplyConfiguration) WithValue(value projectcalicov3.BGPCommunityValue) *BGPFilterAddCommunityApplyConfiguration {
	b.Value = &value
	return b
}
