// Copyright (c) 2020 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

//go:build !windows

package timeshim

import (
	"github.com/projectcalico/calico/felix/bpf"
)

// Time is our shim interface to the time package.
type Interface interface {
	interfaceCommon
	KTimeNanos() int64
}

func (realTime) KTimeNanos() int64 {
	return bpf.KTimeNanos()
}
