/**
 * Link colour tokens
 *
 * Reference: https://designsystem.backbase.com/latest/design-tokens/semantic-colors/link-colors-Yd7FLT6y
 */

export default {
    'experimental-token-link': {
        _light: 'experimental-color-blue.900',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-link:hovered': {
        _light: 'experimental-color-medium-blue.700',
        _dark: 'experimental-color-medium-gold.300',
    },
    'experimental-token-link:pressed': {
        _light: 'experimental-color-medium-blue.600',
        _dark: 'experimental-color-medium-gold.200',
    },
};
