// Code generated by mockery; DO NOT EDIT.
// github.com/vektra/mockery
// template: testify

package mocks

import (
	"context"

	mock "github.com/stretchr/testify/mock"

	"github.com/projectcalico/calico/goldmane/proto"
)

// NewFlowsClient creates a new instance of FlowsClient. It also registers a testing interface on the mock and a cleanup function to assert the mocks expectations.
// The first argument is typically a *testing.T value.
func NewFlowsClient(t interface {
	mock.TestingT
	Cleanup(func())
}) *FlowsClient {
	mock := &FlowsClient{}
	mock.Mock.Test(t)

	t.Cleanup(func() { mock.AssertExpectations(t) })

	return mock
}

// FlowsClient is an autogenerated mock type for the FlowsClient type
type FlowsClient struct {
	mock.Mock
}

type FlowsClient_Expecter struct {
	mock *mock.Mock
}

func (_m *FlowsClient) EXPECT() *FlowsClient_Expecter {
	return &FlowsClient_Expecter{mock: &_m.Mock}
}

// FilterHints provides a mock function for the type FlowsClient
func (_mock *FlowsClient) FilterHints(ctx context.Context, req *proto.FilterHintsRequest) (*proto.ListMetadata, []*proto.FilterHint, error) {
	ret := _mock.Called(ctx, req)

	if len(ret) == 0 {
		panic("no return value specified for FilterHints")
	}

	var r0 *proto.ListMetadata
	var r1 []*proto.FilterHint
	var r2 error
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FilterHintsRequest) (*proto.ListMetadata, []*proto.FilterHint, error)); ok {
		return returnFunc(ctx, req)
	}
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FilterHintsRequest) *proto.ListMetadata); ok {
		r0 = returnFunc(ctx, req)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(*proto.ListMetadata)
		}
	}
	if returnFunc, ok := ret.Get(1).(func(context.Context, *proto.FilterHintsRequest) []*proto.FilterHint); ok {
		r1 = returnFunc(ctx, req)
	} else {
		if ret.Get(1) != nil {
			r1 = ret.Get(1).([]*proto.FilterHint)
		}
	}
	if returnFunc, ok := ret.Get(2).(func(context.Context, *proto.FilterHintsRequest) error); ok {
		r2 = returnFunc(ctx, req)
	} else {
		r2 = ret.Error(2)
	}
	return r0, r1, r2
}

// FlowsClient_FilterHints_Call is a *mock.Call that shadows Run/Return methods with type explicit version for method 'FilterHints'
type FlowsClient_FilterHints_Call struct {
	*mock.Call
}

// FilterHints is a helper method to define mock.On call
//   - ctx context.Context
//   - req *proto.FilterHintsRequest
func (_e *FlowsClient_Expecter) FilterHints(ctx interface{}, req interface{}) *FlowsClient_FilterHints_Call {
	return &FlowsClient_FilterHints_Call{Call: _e.mock.On("FilterHints", ctx, req)}
}

func (_c *FlowsClient_FilterHints_Call) Run(run func(ctx context.Context, req *proto.FilterHintsRequest)) *FlowsClient_FilterHints_Call {
	_c.Call.Run(func(args mock.Arguments) {
		var arg0 context.Context
		if args[0] != nil {
			arg0 = args[0].(context.Context)
		}
		var arg1 *proto.FilterHintsRequest
		if args[1] != nil {
			arg1 = args[1].(*proto.FilterHintsRequest)
		}
		run(
			arg0,
			arg1,
		)
	})
	return _c
}

func (_c *FlowsClient_FilterHints_Call) Return(listMetadata *proto.ListMetadata, filterHints []*proto.FilterHint, err error) *FlowsClient_FilterHints_Call {
	_c.Call.Return(listMetadata, filterHints, err)
	return _c
}

func (_c *FlowsClient_FilterHints_Call) RunAndReturn(run func(ctx context.Context, req *proto.FilterHintsRequest) (*proto.ListMetadata, []*proto.FilterHint, error)) *FlowsClient_FilterHints_Call {
	_c.Call.Return(run)
	return _c
}

// List provides a mock function for the type FlowsClient
func (_mock *FlowsClient) List(context1 context.Context, flowListRequest *proto.FlowListRequest) (*proto.ListMetadata, []*proto.FlowResult, error) {
	ret := _mock.Called(context1, flowListRequest)

	if len(ret) == 0 {
		panic("no return value specified for List")
	}

	var r0 *proto.ListMetadata
	var r1 []*proto.FlowResult
	var r2 error
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FlowListRequest) (*proto.ListMetadata, []*proto.FlowResult, error)); ok {
		return returnFunc(context1, flowListRequest)
	}
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FlowListRequest) *proto.ListMetadata); ok {
		r0 = returnFunc(context1, flowListRequest)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(*proto.ListMetadata)
		}
	}
	if returnFunc, ok := ret.Get(1).(func(context.Context, *proto.FlowListRequest) []*proto.FlowResult); ok {
		r1 = returnFunc(context1, flowListRequest)
	} else {
		if ret.Get(1) != nil {
			r1 = ret.Get(1).([]*proto.FlowResult)
		}
	}
	if returnFunc, ok := ret.Get(2).(func(context.Context, *proto.FlowListRequest) error); ok {
		r2 = returnFunc(context1, flowListRequest)
	} else {
		r2 = ret.Error(2)
	}
	return r0, r1, r2
}

// FlowsClient_List_Call is a *mock.Call that shadows Run/Return methods with type explicit version for method 'List'
type FlowsClient_List_Call struct {
	*mock.Call
}

// List is a helper method to define mock.On call
//   - context1 context.Context
//   - flowListRequest *proto.FlowListRequest
func (_e *FlowsClient_Expecter) List(context1 interface{}, flowListRequest interface{}) *FlowsClient_List_Call {
	return &FlowsClient_List_Call{Call: _e.mock.On("List", context1, flowListRequest)}
}

func (_c *FlowsClient_List_Call) Run(run func(context1 context.Context, flowListRequest *proto.FlowListRequest)) *FlowsClient_List_Call {
	_c.Call.Run(func(args mock.Arguments) {
		var arg0 context.Context
		if args[0] != nil {
			arg0 = args[0].(context.Context)
		}
		var arg1 *proto.FlowListRequest
		if args[1] != nil {
			arg1 = args[1].(*proto.FlowListRequest)
		}
		run(
			arg0,
			arg1,
		)
	})
	return _c
}

func (_c *FlowsClient_List_Call) Return(listMetadata *proto.ListMetadata, flowResults []*proto.FlowResult, err error) *FlowsClient_List_Call {
	_c.Call.Return(listMetadata, flowResults, err)
	return _c
}

func (_c *FlowsClient_List_Call) RunAndReturn(run func(context1 context.Context, flowListRequest *proto.FlowListRequest) (*proto.ListMetadata, []*proto.FlowResult, error)) *FlowsClient_List_Call {
	_c.Call.Return(run)
	return _c
}

// Stream provides a mock function for the type FlowsClient
func (_mock *FlowsClient) Stream(ctx context.Context, request *proto.FlowStreamRequest) (proto.Flows_StreamClient, error) {
	ret := _mock.Called(ctx, request)

	if len(ret) == 0 {
		panic("no return value specified for Stream")
	}

	var r0 proto.Flows_StreamClient
	var r1 error
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FlowStreamRequest) (proto.Flows_StreamClient, error)); ok {
		return returnFunc(ctx, request)
	}
	if returnFunc, ok := ret.Get(0).(func(context.Context, *proto.FlowStreamRequest) proto.Flows_StreamClient); ok {
		r0 = returnFunc(ctx, request)
	} else {
		if ret.Get(0) != nil {
			r0 = ret.Get(0).(proto.Flows_StreamClient)
		}
	}
	if returnFunc, ok := ret.Get(1).(func(context.Context, *proto.FlowStreamRequest) error); ok {
		r1 = returnFunc(ctx, request)
	} else {
		r1 = ret.Error(1)
	}
	return r0, r1
}

// FlowsClient_Stream_Call is a *mock.Call that shadows Run/Return methods with type explicit version for method 'Stream'
type FlowsClient_Stream_Call struct {
	*mock.Call
}

// Stream is a helper method to define mock.On call
//   - ctx context.Context
//   - request *proto.FlowStreamRequest
func (_e *FlowsClient_Expecter) Stream(ctx interface{}, request interface{}) *FlowsClient_Stream_Call {
	return &FlowsClient_Stream_Call{Call: _e.mock.On("Stream", ctx, request)}
}

func (_c *FlowsClient_Stream_Call) Run(run func(ctx context.Context, request *proto.FlowStreamRequest)) *FlowsClient_Stream_Call {
	_c.Call.Run(func(args mock.Arguments) {
		var arg0 context.Context
		if args[0] != nil {
			arg0 = args[0].(context.Context)
		}
		var arg1 *proto.FlowStreamRequest
		if args[1] != nil {
			arg1 = args[1].(*proto.FlowStreamRequest)
		}
		run(
			arg0,
			arg1,
		)
	})
	return _c
}

func (_c *FlowsClient_Stream_Call) Return(v proto.Flows_StreamClient, err error) *FlowsClient_Stream_Call {
	_c.Call.Return(v, err)
	return _c
}

func (_c *FlowsClient_Stream_Call) RunAndReturn(run func(ctx context.Context, request *proto.FlowStreamRequest) (proto.Flows_StreamClient, error)) *FlowsClient_Stream_Call {
	_c.Call.Return(run)
	return _c
}
