// Code generated by "stringer -type=RouteClass"; DO NOT EDIT.

package routetable

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[RouteClassLocalWorkload-0]
	_ = x[RouteClassBPFSpecial-1]
	_ = x[RouteClassWireguard-2]
	_ = x[RouteClassVXLANSameSubnet-3]
	_ = x[RouteClassVXLANTunnel-4]
	_ = x[RouteClassIPIPSameSubnet-5]
	_ = x[RouteClassIPIPTunnel-6]
	_ = x[RouteClassNoEncap-7]
	_ = x[RouteClassIPAMBlockDrop-8]
	_ = x[RouteClassMax-9]
}

const _RouteClass_name = "RouteClassLocalWorkloadRouteClassBPFSpecialRouteClassWireguardRouteClassVXLANSameSubnetRouteClassVXLANTunnelRouteClassIPIPSameSubnetRouteClassIPIPTunnelRouteClassNoEncapRouteClassIPAMBlockDropRouteClassMax"

var _RouteClass_index = [...]uint8{0, 23, 43, 62, 87, 108, 132, 152, 169, 192, 205}

func (i RouteClass) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_RouteClass_index)-1 {
		return "RouteClass(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _RouteClass_name[_RouteClass_index[idx]:_RouteClass_index[idx+1]]
}
