// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

import (
	v1 "k8s.io/client-go/applyconfigurations/meta/v1"
)

// TierStatusApplyConfiguration represents a declarative configuration of the TierStatus type for use
// with apply.
//
// TierStatus contains the status of a Tier resource.
type TierStatusApplyConfiguration struct {
	// Conditions represents the latest observed set of conditions for the resource. A tier with a
	// "Ready" condition set to "True" is operating as expected.
	Conditions []v1.ConditionApplyConfiguration `json:"conditions,omitempty"`
}

// TierStatusApplyConfiguration constructs a declarative configuration of the TierStatus type for use with
// apply.
func TierStatus() *TierStatusApplyConfiguration {
	return &TierStatusApplyConfiguration{}
}

// WithConditions adds the given value to the Conditions field in the declarative configuration
// and returns the receiver, so that objects can be build by chaining "With" function invocations.
// If called multiple times, values provided by each call will be appended to the Conditions field.
func (b *TierStatusApplyConfiguration) WithConditions(values ...*v1.ConditionApplyConfiguration) *TierStatusApplyConfiguration {
	for i := range values {
		if values[i] == nil {
			panic("nil value passed to WithConditions")
		}
		b.Conditions = append(b.Conditions, *values[i])
	}
	return b
}
