// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by client-gen. DO NOT EDIT.

package fake

import (
	v3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	projectcalicov3 "github.com/projectcalico/api/pkg/client/applyconfiguration_generated/projectcalico/v3"
	typedprojectcalicov3 "github.com/projectcalico/api/pkg/client/clientset_generated/clientset/typed/projectcalico/v3"
	gentype "k8s.io/client-go/gentype"
)

// fakeHostEndpoints implements HostEndpointInterface
type fakeHostEndpoints struct {
	*gentype.FakeClientWithListAndApply[*v3.HostEndpoint, *v3.HostEndpointList, *projectcalicov3.HostEndpointApplyConfiguration]
	Fake *FakeProjectcalicoV3
}

func newFakeHostEndpoints(fake *FakeProjectcalicoV3) typedprojectcalicov3.HostEndpointInterface {
	return &fakeHostEndpoints{
		gentype.NewFakeClientWithListAndApply[*v3.HostEndpoint, *v3.HostEndpointList, *projectcalicov3.HostEndpointApplyConfiguration](
			fake.Fake,
			"",
			v3.SchemeGroupVersion.WithResource("hostendpoints"),
			v3.SchemeGroupVersion.WithKind("HostEndpoint"),
			func() *v3.HostEndpoint { return &v3.HostEndpoint{} },
			func() *v3.HostEndpointList { return &v3.HostEndpointList{} },
			func(dst, src *v3.HostEndpointList) { dst.ListMeta = src.ListMeta },
			func(list *v3.HostEndpointList) []*v3.HostEndpoint { return gentype.ToPointerSlice(list.Items) },
			func(list *v3.HostEndpointList, items []*v3.HostEndpoint) {
				list.Items = gentype.FromPointerSlice(items)
			},
		),
		fake,
	}
}
