// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.8
// 	protoc        v6.32.0
// source: healthz.proto

package proto

import (
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"

	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type HealthCheckRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *HealthCheckRequest) Reset() {
	*x = HealthCheckRequest{}
	mi := &file_healthz_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *HealthCheckRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*HealthCheckRequest) ProtoMessage() {}

func (x *HealthCheckRequest) ProtoReflect() protoreflect.Message {
	mi := &file_healthz_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use HealthCheckRequest.ProtoReflect.Descriptor instead.
func (*HealthCheckRequest) Descriptor() ([]byte, []int) {
	return file_healthz_proto_rawDescGZIP(), []int{0}
}

type HealthCheckResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Healthy       bool                   `protobuf:"varint,1,opt,name=healthy,proto3" json:"healthy,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *HealthCheckResponse) Reset() {
	*x = HealthCheckResponse{}
	mi := &file_healthz_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *HealthCheckResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*HealthCheckResponse) ProtoMessage() {}

func (x *HealthCheckResponse) ProtoReflect() protoreflect.Message {
	mi := &file_healthz_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use HealthCheckResponse.ProtoReflect.Descriptor instead.
func (*HealthCheckResponse) Descriptor() ([]byte, []int) {
	return file_healthz_proto_rawDescGZIP(), []int{1}
}

func (x *HealthCheckResponse) GetHealthy() bool {
	if x != nil {
		return x.Healthy
	}
	return false
}

var File_healthz_proto protoreflect.FileDescriptor

const file_healthz_proto_rawDesc = "" +
	"\n" +
	"\rhealthz.proto\x12\bdikastes\"\x14\n" +
	"\x12HealthCheckRequest\"/\n" +
	"\x13HealthCheckResponse\x12\x18\n" +
	"\ahealthy\x18\x01 \x01(\bR\ahealthy2\xa6\x01\n" +
	"\aHealthz\x12M\n" +
	"\x0eCheckReadiness\x12\x1c.dikastes.HealthCheckRequest\x1a\x1d.dikastes.HealthCheckResponse\x12L\n" +
	"\rCheckLiveness\x12\x1c.dikastes.HealthCheckRequest\x1a\x1d.dikastes.HealthCheckResponseB\tZ\a./protob\x06proto3"

var (
	file_healthz_proto_rawDescOnce sync.Once
	file_healthz_proto_rawDescData []byte
)

func file_healthz_proto_rawDescGZIP() []byte {
	file_healthz_proto_rawDescOnce.Do(func() {
		file_healthz_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_healthz_proto_rawDesc), len(file_healthz_proto_rawDesc)))
	})
	return file_healthz_proto_rawDescData
}

var file_healthz_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_healthz_proto_goTypes = []any{
	(*HealthCheckRequest)(nil),  // 0: dikastes.HealthCheckRequest
	(*HealthCheckResponse)(nil), // 1: dikastes.HealthCheckResponse
}
var file_healthz_proto_depIdxs = []int32{
	0, // 0: dikastes.Healthz.CheckReadiness:input_type -> dikastes.HealthCheckRequest
	0, // 1: dikastes.Healthz.CheckLiveness:input_type -> dikastes.HealthCheckRequest
	1, // 2: dikastes.Healthz.CheckReadiness:output_type -> dikastes.HealthCheckResponse
	1, // 3: dikastes.Healthz.CheckLiveness:output_type -> dikastes.HealthCheckResponse
	2, // [2:4] is the sub-list for method output_type
	0, // [0:2] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_healthz_proto_init() }
func file_healthz_proto_init() {
	if File_healthz_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_healthz_proto_rawDesc), len(file_healthz_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_healthz_proto_goTypes,
		DependencyIndexes: file_healthz_proto_depIdxs,
		MessageInfos:      file_healthz_proto_msgTypes,
	}.Build()
	File_healthz_proto = out.File
	file_healthz_proto_goTypes = nil
	file_healthz_proto_depIdxs = nil
}
