// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by lister-gen. DO NOT EDIT.

package v3

// BGPConfigurationListerExpansion allows custom methods to be added to
// BGPConfigurationLister.
type BGPConfigurationListerExpansion interface{}

// BGPFilterListerExpansion allows custom methods to be added to
// BGPFilterLister.
type BGPFilterListerExpansion interface{}

// BGPPeerListerExpansion allows custom methods to be added to
// BGPPeerLister.
type BGPPeerListerExpansion interface{}

// BlockAffinityListerExpansion allows custom methods to be added to
// BlockAffinityLister.
type BlockAffinityListerExpansion interface{}

// CalicoNodeStatusListerExpansion allows custom methods to be added to
// CalicoNodeStatusLister.
type CalicoNodeStatusListerExpansion interface{}

// ClusterInformationListerExpansion allows custom methods to be added to
// ClusterInformationLister.
type ClusterInformationListerExpansion interface{}

// FelixConfigurationListerExpansion allows custom methods to be added to
// FelixConfigurationLister.
type FelixConfigurationListerExpansion interface{}

// GlobalNetworkPolicyListerExpansion allows custom methods to be added to
// GlobalNetworkPolicyLister.
type GlobalNetworkPolicyListerExpansion interface{}

// GlobalNetworkSetListerExpansion allows custom methods to be added to
// GlobalNetworkSetLister.
type GlobalNetworkSetListerExpansion interface{}

// HostEndpointListerExpansion allows custom methods to be added to
// HostEndpointLister.
type HostEndpointListerExpansion interface{}

// IPAMBlockListerExpansion allows custom methods to be added to
// IPAMBlockLister.
type IPAMBlockListerExpansion interface{}

// IPAMConfigurationListerExpansion allows custom methods to be added to
// IPAMConfigurationLister.
type IPAMConfigurationListerExpansion interface{}

// IPAMHandleListerExpansion allows custom methods to be added to
// IPAMHandleLister.
type IPAMHandleListerExpansion interface{}

// IPAMHandleNamespaceListerExpansion allows custom methods to be added to
// IPAMHandleNamespaceLister.
type IPAMHandleNamespaceListerExpansion interface{}

// IPPoolListerExpansion allows custom methods to be added to
// IPPoolLister.
type IPPoolListerExpansion interface{}

// IPReservationListerExpansion allows custom methods to be added to
// IPReservationLister.
type IPReservationListerExpansion interface{}

// KubeControllersConfigurationListerExpansion allows custom methods to be added to
// KubeControllersConfigurationLister.
type KubeControllersConfigurationListerExpansion interface{}

// NetworkPolicyListerExpansion allows custom methods to be added to
// NetworkPolicyLister.
type NetworkPolicyListerExpansion interface{}

// NetworkPolicyNamespaceListerExpansion allows custom methods to be added to
// NetworkPolicyNamespaceLister.
type NetworkPolicyNamespaceListerExpansion interface{}

// NetworkSetListerExpansion allows custom methods to be added to
// NetworkSetLister.
type NetworkSetListerExpansion interface{}

// NetworkSetNamespaceListerExpansion allows custom methods to be added to
// NetworkSetNamespaceLister.
type NetworkSetNamespaceListerExpansion interface{}

// ProfileListerExpansion allows custom methods to be added to
// ProfileLister.
type ProfileListerExpansion interface{}

// StagedGlobalNetworkPolicyListerExpansion allows custom methods to be added to
// StagedGlobalNetworkPolicyLister.
type StagedGlobalNetworkPolicyListerExpansion interface{}

// StagedKubernetesNetworkPolicyListerExpansion allows custom methods to be added to
// StagedKubernetesNetworkPolicyLister.
type StagedKubernetesNetworkPolicyListerExpansion interface{}

// StagedKubernetesNetworkPolicyNamespaceListerExpansion allows custom methods to be added to
// StagedKubernetesNetworkPolicyNamespaceLister.
type StagedKubernetesNetworkPolicyNamespaceListerExpansion interface{}

// StagedNetworkPolicyListerExpansion allows custom methods to be added to
// StagedNetworkPolicyLister.
type StagedNetworkPolicyListerExpansion interface{}

// StagedNetworkPolicyNamespaceListerExpansion allows custom methods to be added to
// StagedNetworkPolicyNamespaceLister.
type StagedNetworkPolicyNamespaceListerExpansion interface{}

// TierListerExpansion allows custom methods to be added to
// TierLister.
type TierListerExpansion interface{}
