// Copyright (c) 2017 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package stringutils

func CommonPrefix(strs []string) string {
	if len(strs) == 0 {
		return ""
	}
	candidatePrefix := strs[0]
	for _, s := range strs[1:] {
		candidatePrefix = commonPrefixTwoStrings(candidatePrefix, s)
	}
	return candidatePrefix
}

func commonPrefixTwoStrings(a, b string) string {
	commonLen := 0
	for ; commonLen < len(a) && commonLen < len(b); commonLen++ {
		if a[commonLen] != b[commonLen] {
			break
		}
	}
	return a[:commonLen]
}
