syntax = "proto3";
package udsver.v1;
option go_package = "./proto";

service Verify {
	rpc Check(Request) returns (Response);
}

message Request {
	string name = 1;
}

message Response {
	// TODO Maybe replace with googleapis
	message Status {
		enum Code {
			// https://cloud.google.com/appengine/docs/admin-api/reference/rpc/google.rpc#google.rpc.Code
			OK = 0;
			CANCELLED = 1;
			UNKNOWN = 2;
			INVALID_ARGUMENT = 3;
			DEADLINE_EXCEEDED = 4;
			NOT_FOUND = 5;
			ALREADY_EXISTS = 6;
			PERMISSION_DENIED = 7;
			UNAUTHENTICATED = 8;
			RESOURCE_EXHAUSTED = 9;
			FAILED_PRECONDITION = 10;
			ABORTED = 11;
			OUT_OF_RANGE = 12;
			UNIMPLEMENTED = 13;
			INTERNAL = 14;
			UNAVAILABLE = 15;
			DATA_LOSS = 16;
		}
		Code code = 1;
		string message = 2;
	}

	Status status = 1;
}
