export const noResultsStyles = {
    display: 'block',
    div: {
        textAlign: 'center',
        display: 'block',
        padding: '8',
        color: 'tigera-color-on-surface',
    },
    bg: 'tigera-color-surface-secondary',
};
