export default {
    'html, body': {
        bg: 'tigera-color-surface',
        color: 'tigera-color-on-surface',
        fontFamily:
            '"Poppins", "Helvetica Neue", "Helvetica", "Arial", sans-serif',
    },
};
