/**
 * On-background colours are contrast paired tokens for background colours to insure an acceptable contrast ratio.
 *
 * Reference: https://designsystem.backbase.com/latest/design-tokens/semantic-colors/on-background-colors-aJ93P0Po
 */

export default {
    'experimental-token-on-bg-brand': {
        _light: 'experimental-token-fg-inverted',
        _dark: 'experimental-token-fg-inverted',
    },
    'experimental-token-on-bg-warning': {
        _light: 'experimental-token-fg-inverted',
        _dark: 'experimental-token-fg-inverted',
    },
    'experimental-token-on-bg-warning-subtle': {
        _light: 'experimental-color-medium-gold.1000',
        _dark: 'experimental-color-medium-gold.50',
    },
    'experimental-token-on-bg-info': {
        _light: 'experimental-token-fg-inverted',
        _dark: 'experimental-token-fg-inverted',
    },
    'experimental-token-on-bg-info-subtle': {
        _light: 'experimental-token-fg-default',
        _dark: 'experimental-color-medium-blue.100',
    },
    'experimental-token-on-bg-danger': {
        _light: 'experimental-token-fg-inverted',
        _dark: 'experimental-token-fg-inverted',
    },
    'experimental-token-on-bg-danger-subtle': {
        _light: 'experimental-color-red.1000',
        _dark: 'experimental-color-red.200',
    },
    'experimental-token-on-bg-success': {
        _light: 'experimental-color-neutral.1200',
        _dark: 'experimental-color-neutral.1200',
    },
    'experimental-token-on-bg-success-subtle': {
        _light: 'experimental-color-green.1000',
        _dark: 'experimental-color-green.200',
    },
};
