export const tableStyles = {
    fontFamily: 'monospace, monospace',
    overflow: 'hidden',
};

export const headerStyles = {
    '>div>div': {
        textTransform: 'none',
        py: '2px',
    },
    '>div>div:last-of-type': {
        py: 0,
    },
    position: 'sticky',
    top: 0,
    zIndex: 1,
};

export const subRowStyles = {
    bg: 'tigera-color-surface',
};
