// Copyright (c) 2020-2022 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//	http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package wireguard

import "time"

type Config struct {
	// Wireguard configuration
	Enabled             bool
	EnabledV6           bool
	ListeningPort       int
	ListeningPortV6     int
	FirewallMark        int
	RoutingRulePriority int
	RoutingTableIndex   int
	RoutingTableIndexV6 int
	InterfaceName       string
	InterfaceNameV6     string
	MTU                 int
	MTUV6               int
	RouteSource         string
	EncryptHostTraffic  bool
	PersistentKeepAlive time.Duration
	RouteSyncDisabled   bool
	ThreadedNAPI        bool
}
