export const iconButtonStyles = {
    fontSize: '2xs',
    color: 'tigeraGrey.400',
    _hover: {
        color: 'tigeraGrey.200',
    },
};

export const labelStyles = {
    fontSize: 'sm',
    display: 'flex',
    alignItems: 'center',
    gap: 1,
};
