export const containerStyles = {
    justifyContent: 'space-between',
    minH: '38px',
    border: '1px solid',
    borderColor: 'tigeraGrey.800',
    borderRadius: 'md',
    pl: 2,
    alignItems: 'center',
    display: 'flex',
    width: '100%',
};

export const iconButtonStyles = {
    fontSize: 'lg',
    minW: 8,
    flexShrink: 0,
};

export const iconStyles = {
    color: 'tigeraGrey.400',
    transition: '100ms',
    _hover: {
        color: 'tigeraGrey.50',
    },
};
