// Copyright (c) 2024 Tigera, Inc. All rights reserved.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package command

// Make runs a make command.
func Make(args []string, envs []string) (string, error) {
	return runner().Run("make", args, envs)
}

// MakeInDir runs a make command in a specific directory.
func MakeInDir(dir string, args []string, envs []string) (string, error) {
	args = append([]string{"-C", dir}, args...)
	return Make(args, envs)
}
