// Copyright (c) 2021 Tigera, Inc. All rights reserved.

// +k8s:deepcopy-gen=package,register
// +k8s:openapi-gen=true
// +k8s:openapi-model-package=com.github.projectcalico.api.pkg.apis.projectcalico.v3

// Package v3 is the v3 version of the API.
// +groupName=projectcalico.org

package v3 // import "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
