export const appHeaderStyles = {
    bg: 'tigera-bg',
    minHeight: 12,
    px: 4,
    py: 2,
    alignItems: 'center',
    justifyContent: 'space-between',
    borderBottom: '2px solid',
    borderBottomColor: 'tigera-color-outline',
};
