// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by client-gen. DO NOT EDIT.

package v3

import (
	context "context"

	projectcalicov3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	applyconfigurationgeneratedprojectcalicov3 "github.com/projectcalico/api/pkg/client/applyconfiguration_generated/projectcalico/v3"
	scheme "github.com/projectcalico/api/pkg/client/clientset_generated/clientset/scheme"
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	types "k8s.io/apimachinery/pkg/types"
	watch "k8s.io/apimachinery/pkg/watch"
	gentype "k8s.io/client-go/gentype"
)

// BGPFiltersGetter has a method to return a BGPFilterInterface.
// A group's client should implement this interface.
type BGPFiltersGetter interface {
	BGPFilters() BGPFilterInterface
}

// BGPFilterInterface has methods to work with BGPFilter resources.
type BGPFilterInterface interface {
	Create(ctx context.Context, bGPFilter *projectcalicov3.BGPFilter, opts v1.CreateOptions) (*projectcalicov3.BGPFilter, error)
	Update(ctx context.Context, bGPFilter *projectcalicov3.BGPFilter, opts v1.UpdateOptions) (*projectcalicov3.BGPFilter, error)
	Delete(ctx context.Context, name string, opts v1.DeleteOptions) error
	DeleteCollection(ctx context.Context, opts v1.DeleteOptions, listOpts v1.ListOptions) error
	Get(ctx context.Context, name string, opts v1.GetOptions) (*projectcalicov3.BGPFilter, error)
	List(ctx context.Context, opts v1.ListOptions) (*projectcalicov3.BGPFilterList, error)
	Watch(ctx context.Context, opts v1.ListOptions) (watch.Interface, error)
	Patch(ctx context.Context, name string, pt types.PatchType, data []byte, opts v1.PatchOptions, subresources ...string) (result *projectcalicov3.BGPFilter, err error)
	Apply(ctx context.Context, bGPFilter *applyconfigurationgeneratedprojectcalicov3.BGPFilterApplyConfiguration, opts v1.ApplyOptions) (result *projectcalicov3.BGPFilter, err error)
	BGPFilterExpansion
}

// bGPFilters implements BGPFilterInterface
type bGPFilters struct {
	*gentype.ClientWithListAndApply[*projectcalicov3.BGPFilter, *projectcalicov3.BGPFilterList, *applyconfigurationgeneratedprojectcalicov3.BGPFilterApplyConfiguration]
}

// newBGPFilters returns a BGPFilters
func newBGPFilters(c *ProjectcalicoV3Client) *bGPFilters {
	return &bGPFilters{
		gentype.NewClientWithListAndApply[*projectcalicov3.BGPFilter, *projectcalicov3.BGPFilterList, *applyconfigurationgeneratedprojectcalicov3.BGPFilterApplyConfiguration](
			"bgpfilters",
			c.RESTClient(),
			scheme.ParameterCodec,
			"",
			func() *projectcalicov3.BGPFilter { return &projectcalicov3.BGPFilter{} },
			func() *projectcalicov3.BGPFilterList { return &projectcalicov3.BGPFilterList{} },
		),
	}
}
