// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by lister-gen. DO NOT EDIT.

package v3

import (
	projectcalicov3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	labels "k8s.io/apimachinery/pkg/labels"
	listers "k8s.io/client-go/listers"
	cache "k8s.io/client-go/tools/cache"
)

// IPAMBlockLister helps list IPAMBlocks.
// All objects returned here must be treated as read-only.
type IPAMBlockLister interface {
	// List lists all IPAMBlocks in the indexer.
	// Objects returned here must be treated as read-only.
	List(selector labels.Selector) (ret []*projectcalicov3.IPAMBlock, err error)
	// Get retrieves the IPAMBlock from the index for a given name.
	// Objects returned here must be treated as read-only.
	Get(name string) (*projectcalicov3.IPAMBlock, error)
	IPAMBlockListerExpansion
}

// iPAMBlockLister implements the IPAMBlockLister interface.
type iPAMBlockLister struct {
	listers.ResourceIndexer[*projectcalicov3.IPAMBlock]
}

// NewIPAMBlockLister returns a new IPAMBlockLister.
func NewIPAMBlockLister(indexer cache.Indexer) IPAMBlockLister {
	return &iPAMBlockLister{listers.New[*projectcalicov3.IPAMBlock](indexer, projectcalicov3.Resource("ipamblock"))}
}
