// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by lister-gen. DO NOT EDIT.

package v3

import (
	projectcalicov3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	labels "k8s.io/apimachinery/pkg/labels"
	listers "k8s.io/client-go/listers"
	cache "k8s.io/client-go/tools/cache"
)

// IPAMHandleLister helps list IPAMHandles.
// All objects returned here must be treated as read-only.
type IPAMHandleLister interface {
	// List lists all IPAMHandles in the indexer.
	// Objects returned here must be treated as read-only.
	List(selector labels.Selector) (ret []*projectcalicov3.IPAMHandle, err error)
	// IPAMHandles returns an object that can list and get IPAMHandles.
	IPAMHandles(namespace string) IPAMHandleNamespaceLister
	IPAMHandleListerExpansion
}

// iPAMHandleLister implements the IPAMHandleLister interface.
type iPAMHandleLister struct {
	listers.ResourceIndexer[*projectcalicov3.IPAMHandle]
}

// NewIPAMHandleLister returns a new IPAMHandleLister.
func NewIPAMHandleLister(indexer cache.Indexer) IPAMHandleLister {
	return &iPAMHandleLister{listers.New[*projectcalicov3.IPAMHandle](indexer, projectcalicov3.Resource("ipamhandle"))}
}

// IPAMHandles returns an object that can list and get IPAMHandles.
func (s *iPAMHandleLister) IPAMHandles(namespace string) IPAMHandleNamespaceLister {
	return iPAMHandleNamespaceLister{listers.NewNamespaced[*projectcalicov3.IPAMHandle](s.ResourceIndexer, namespace)}
}

// IPAMHandleNamespaceLister helps list and get IPAMHandles.
// All objects returned here must be treated as read-only.
type IPAMHandleNamespaceLister interface {
	// List lists all IPAMHandles in the indexer for a given namespace.
	// Objects returned here must be treated as read-only.
	List(selector labels.Selector) (ret []*projectcalicov3.IPAMHandle, err error)
	// Get retrieves the IPAMHandle from the indexer for a given namespace and name.
	// Objects returned here must be treated as read-only.
	Get(name string) (*projectcalicov3.IPAMHandle, error)
	IPAMHandleNamespaceListerExpansion
}

// iPAMHandleNamespaceLister implements the IPAMHandleNamespaceLister
// interface.
type iPAMHandleNamespaceLister struct {
	listers.ResourceIndexer[*projectcalicov3.IPAMHandle]
}
