// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by client-gen. DO NOT EDIT.

package v3

type BGPConfigurationExpansion interface{}

type BGPFilterExpansion interface{}

type BGPPeerExpansion interface{}

type BlockAffinityExpansion interface{}

type CalicoNodeStatusExpansion interface{}

type ClusterInformationExpansion interface{}

type FelixConfigurationExpansion interface{}

type GlobalNetworkPolicyExpansion interface{}

type GlobalNetworkSetExpansion interface{}

type HostEndpointExpansion interface{}

type IPAMBlockExpansion interface{}

type IPAMConfigurationExpansion interface{}

type IPAMHandleExpansion interface{}

type IPPoolExpansion interface{}

type IPReservationExpansion interface{}

type KubeControllersConfigurationExpansion interface{}

type NetworkPolicyExpansion interface{}

type NetworkSetExpansion interface{}

type ProfileExpansion interface{}

type StagedGlobalNetworkPolicyExpansion interface{}

type StagedKubernetesNetworkPolicyExpansion interface{}

type StagedNetworkPolicyExpansion interface{}

type TierExpansion interface{}
