export const listContainerStyles = {
    mx: 4,
    my: 2,
    flexDirection: 'column',
    gap: 2,
};

export const switchStyles = {
    py: 1,
    display: 'flex',
    alignItems: 'center',
    fontWeight: 'normal',
};
