//go:build !ignore_autogenerated
// +build !ignore_autogenerated

// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by openapi-gen. DO NOT EDIT.

package v3

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in AllocationAttribute) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.AllocationAttribute"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in AutoHostEndpointConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.AutoHostEndpointConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPConfiguration) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPConfiguration"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPConfigurationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPConfigurationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPConfigurationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPConfigurationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPDaemonStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPDaemonStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilter) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilter"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterAddCommunity) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterAddCommunity"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterCommunityMatch) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterCommunityMatch"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterOperation) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterOperation"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterPrefixLengthV4) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterPrefixLengthV4"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterPrefixLengthV6) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterPrefixLengthV6"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterPrependASPath) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterPrependASPath"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterRuleV4) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterRuleV4"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterRuleV6) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterRuleV6"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterSetPriority) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterSetPriority"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPFilterSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPFilterSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPPassword) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPPassword"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPPeer) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPPeer"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPPeerList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPPeerList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BGPPeerSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BGPPeerSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BPFConntrackTimeouts) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BPFConntrackTimeouts"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BlockAffinity) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BlockAffinity"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BlockAffinityList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BlockAffinityList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in BlockAffinitySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.BlockAffinitySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeAgentStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeAgentStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeBGPRouteStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeBGPRouteStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeBGPStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeBGPStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodePeer) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodePeer"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeRoute) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeRoute"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeRouteLearnedFrom) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeRouteLearnedFrom"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeStatusList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeStatusList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeStatusSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeStatusSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in CalicoNodeStatusStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.CalicoNodeStatusStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ClusterInformation) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ClusterInformation"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ClusterInformationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ClusterInformationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ClusterInformationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ClusterInformationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Community) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.Community"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ControllersConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ControllersConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in EndpointPort) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.EndpointPort"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in EntityRule) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.EntityRule"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in FelixConfiguration) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.FelixConfiguration"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in FelixConfigurationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.FelixConfigurationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in FelixConfigurationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.FelixConfigurationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkPolicy) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkPolicy"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkPolicyList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkPolicyList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkPolicySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkPolicySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkSet) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkSet"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkSetList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkSetList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in GlobalNetworkSetSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.GlobalNetworkSetSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HTTPMatch) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HTTPMatch"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HTTPPath) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HTTPPath"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HealthTimeoutOverride) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HealthTimeoutOverride"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HostEndpoint) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HostEndpoint"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HostEndpointList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HostEndpointList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in HostEndpointSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.HostEndpointSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ICMPFields) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ICMPFields"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMBlock) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMBlock"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMBlockList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMBlockList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMBlockSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMBlockSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMConfiguration) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMConfiguration"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMConfigurationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMConfigurationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMConfigurationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMConfigurationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMHandle) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMHandle"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMHandleList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMHandleList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPAMHandleSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPAMHandleSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPIPConfiguration) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPIPConfiguration"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPPool) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPPool"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPPoolList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPPoolList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPPoolSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPPoolSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPPoolStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPPoolStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPReservation) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPReservation"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPReservationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPReservationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in IPReservationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.IPReservationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in KubeControllersConfiguration) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.KubeControllersConfiguration"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in KubeControllersConfigurationList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.KubeControllersConfigurationList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in KubeControllersConfigurationSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.KubeControllersConfigurationSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in KubeControllersConfigurationStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.KubeControllersConfigurationStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in LoadBalancerControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.LoadBalancerControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in MigrationControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.MigrationControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NamespaceControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NamespaceControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkPolicy) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkPolicy"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkPolicyList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkPolicyList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkPolicySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkPolicySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkSet) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkSet"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkSetList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkSetList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NetworkSetSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NetworkSetSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in NodeControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.NodeControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in PolicyControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.PolicyControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in PrefixAdvertisement) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.PrefixAdvertisement"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Profile) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.Profile"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ProfileList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ProfileList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ProfileSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ProfileSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ProtoPort) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ProtoPort"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in RouteTableIDRange) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.RouteTableIDRange"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in RouteTableRange) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.RouteTableRange"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Rule) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.Rule"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in RuleMetadata) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.RuleMetadata"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceAccountControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceAccountControllerConfig"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceAccountMatch) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceAccountMatch"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceClusterIPBlock) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceClusterIPBlock"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceExternalIPBlock) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceExternalIPBlock"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceLoadBalancerIPBlock) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceLoadBalancerIPBlock"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in ServiceMatch) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.ServiceMatch"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedGlobalNetworkPolicy) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedGlobalNetworkPolicy"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedGlobalNetworkPolicyList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedGlobalNetworkPolicyList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedGlobalNetworkPolicySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedGlobalNetworkPolicySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedKubernetesNetworkPolicy) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedKubernetesNetworkPolicy"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedKubernetesNetworkPolicyList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedKubernetesNetworkPolicyList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedKubernetesNetworkPolicySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedKubernetesNetworkPolicySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedNetworkPolicy) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedNetworkPolicy"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedNetworkPolicyList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedNetworkPolicyList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in StagedNetworkPolicySpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.StagedNetworkPolicySpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Template) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.Template"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in Tier) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.Tier"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in TierList) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.TierList"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in TierSpec) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.TierSpec"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in TierStatus) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.TierStatus"
}

// OpenAPIModelName returns the OpenAPI model name for this type.
func (in WorkloadEndpointControllerConfig) OpenAPIModelName() string {
	return "com.github.projectcalico.api.pkg.apis.projectcalico.v3.WorkloadEndpointControllerConfig"
}
