// Copyright (c) 2022 Tigera, Inc. All rights reserved.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package v3

import (
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

const (
	KindIPAMConfiguration     = "IPAMConfiguration"
	KindIPAMConfigurationList = "IPAMConfigurationList"
)

// +genclient:nonNamespaced
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object
// +kubebuilder:resource:scope=Cluster

// IPAMConfigurationList contains a list of IPAMConfiguration resources.
type IPAMConfigurationList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata" protobuf:"bytes,1,opt,name=metadata"`

	Items []IPAMConfiguration `json:"items" protobuf:"bytes,2,rep,name=items"`
}

// +genclient
// +genclient:nonNamespaced
// +k8s:deepcopy-gen:interfaces=k8s.io/apimachinery/pkg/runtime.Object
// +kubebuilder:resource:scope=Cluster,shortName={ipamconfig,ipamconfigs}
// +kubebuilder:printcolumn:name="StrictAffinity",type=boolean,JSONPath=".spec.strictAffinity",description="Whether borrowing IP addresses is allowed"
// +kubebuilder:printcolumn:name="MaxBlocksPerHost",type=integer,JSONPath=".spec.maxBlocksPerHost",description="The max number of blocks that can be affine to each host"
// +kubebuilder:printcolumn:name="AutoAllocateBlocks",type=boolean,JSONPath=".spec.autoAllocateBlocks",description="Whether or not to auto allocate blocks to hosts"

// IPAMConfiguration contains information about a block for IP address assignment.
type IPAMConfiguration struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata" protobuf:"bytes,1,opt,name=metadata"`

	Spec IPAMConfigurationSpec `json:"spec" protobuf:"bytes,2,opt,name=spec"`
}

// VMAddressPersistence controls whether KubeVirt VirtualMachine workloads
// maintain persistent IP addresses across VM lifecycle events.
type VMAddressPersistence string

const (
	VMAddressPersistenceEnabled  VMAddressPersistence = "Enabled"
	VMAddressPersistenceDisabled VMAddressPersistence = "Disabled"
)

// IPAMConfigurationSpec contains the specification for an IPAMConfiguration resource.
type IPAMConfigurationSpec struct {
	// When StrictAffinity is true, borrowing IP addresses is not allowed.
	// +kubebuilder:default=false
	StrictAffinity bool `json:"strictAffinity"`

	// MaxBlocksPerHost, if non-zero, is the max number of blocks that can be
	// affine to each host.
	//
	// +kubebuilder:validation:Minimum=0
	// +kubebuilder:validation:Maximum=1000000
	// +kubebuilder:default=0
	// +optional
	MaxBlocksPerHost int32 `json:"maxBlocksPerHost,omitempty"`

	// Whether or not to auto allocate blocks to hosts.
	// +kubebuilder:default=true
	AutoAllocateBlocks bool `json:"autoAllocateBlocks"`

	// KubeVirtVMAddressPersistence controls whether KubeVirt VirtualMachine workloads
	// maintain persistent IP addresses across VM lifecycle events (reboot, migration, pod eviction).
	// When Enabled, Calico automatically ensures that KubeVirt VMs retain their IP addresses
	// when their underlying pods are recreated during VM operations.
	// When Disabled, VMs receive new IP addresses whenever their pods are recreated,
	// and creating a live migration target pod is not supported because the migration
	// target pod requires the same IP as the source pod, which is only possible with
	// address persistence.
	// Defaults to Enabled if not specified.
	// +kubebuilder:validation:Enum=Enabled;Disabled
	// +optional
	KubeVirtVMAddressPersistence *VMAddressPersistence `json:"kubeVirtVMAddressPersistence,omitempty"`
}

// NewIPAMConfiguration creates a new (zeroed) IPAMConfiguration struct with the TypeMetadata initialised to the current
// version.
func NewIPAMConfiguration() *IPAMConfiguration {
	return &IPAMConfiguration{
		TypeMeta: metav1.TypeMeta{
			Kind:       KindIPAMConfiguration,
			APIVersion: GroupVersionCurrent,
		},
	}
}
