import { listItemStyles as checkboxListItemStyles } from '../OmniCheckboxList/styles';

export const listItemStyles = {
    ...checkboxListItemStyles,
    borderLeftWidth: 0,
};

export const radioStyles = {
    pt: 1,
};
