/**
 * Elevation tokens are used to represent how containers are layered on top of each other.
 *
 * References:
 * - https://atlassian.design/foundations/elevation
 * - https://designsystem.backbase.com/latest/design-tokens/semantic-colors/background-colors-mLRqePLh#search-242c855f-4858-4e29-b05e-310db7945eebJ
 */

export default {
    'experimental-token-elevation-sunken': {
        _light: 'experimental-color-neutral.50',
        _dark: 'experimental-color-neutral.1200',
    },
    'experimental-token-elevation-displaced': {
        _light: 'experimental-color-light-alpha.500',
        _dark: 'experimental-color-dark-alpha.500',
    },
    'experimental-token-elevation-page': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1100',
    },
    'experimental-token-elevation-surface': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1000',
    },
    'experimental-token-elevation-surface:hovered': {
        _light: 'experimental-color-neutral.50',
        _dark: 'experimental-color-neutral.900',
    },
    'experimental-token-elevation-surface:pressed': {
        _light: 'experimental-color-neutral.100',
        _dark: 'experimental-color-neutral.800',
    },
    'experimental-token-elevation-surface-shadow': {
        _light: '0px 0px 8px rgba(20, 20, 20, 0.05)',
        _dark: '0px 0px 8px rgba(37, 38, 39, 0.1)',
    },
    'experimental-token-elevation-overlay': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1000',
    },
    'experimental-token-elevation-overlay-shadow': {
        _light: '0px 0px 8px rgba(20, 20, 20, 0.1)',
        _dark: '0px 0px 8px rgba(20, 20, 20, 0.5)',
    },
    'experimental-token-elevation-overlay-inverted': {
        _light: 'experimental-color-neutral.1100',
        _dark: 'experimental-color-neutral.50',
    },
    'experimental-token-on-elevation-overlay-inverted': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1100',
    },
    'experimental-token-elevation-overlay-surface': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1100',
    },
};
