// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v6.33.4
// source: udsver.proto

package proto

import (
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"

	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Response_Status_Code int32

const (
	// https://cloud.google.com/appengine/docs/admin-api/reference/rpc/google.rpc#google.rpc.Code
	Response_Status_OK                  Response_Status_Code = 0
	Response_Status_CANCELLED           Response_Status_Code = 1
	Response_Status_UNKNOWN             Response_Status_Code = 2
	Response_Status_INVALID_ARGUMENT    Response_Status_Code = 3
	Response_Status_DEADLINE_EXCEEDED   Response_Status_Code = 4
	Response_Status_NOT_FOUND           Response_Status_Code = 5
	Response_Status_ALREADY_EXISTS      Response_Status_Code = 6
	Response_Status_PERMISSION_DENIED   Response_Status_Code = 7
	Response_Status_UNAUTHENTICATED     Response_Status_Code = 8
	Response_Status_RESOURCE_EXHAUSTED  Response_Status_Code = 9
	Response_Status_FAILED_PRECONDITION Response_Status_Code = 10
	Response_Status_ABORTED             Response_Status_Code = 11
	Response_Status_OUT_OF_RANGE        Response_Status_Code = 12
	Response_Status_UNIMPLEMENTED       Response_Status_Code = 13
	Response_Status_INTERNAL            Response_Status_Code = 14
	Response_Status_UNAVAILABLE         Response_Status_Code = 15
	Response_Status_DATA_LOSS           Response_Status_Code = 16
)

// Enum value maps for Response_Status_Code.
var (
	Response_Status_Code_name = map[int32]string{
		0:  "OK",
		1:  "CANCELLED",
		2:  "UNKNOWN",
		3:  "INVALID_ARGUMENT",
		4:  "DEADLINE_EXCEEDED",
		5:  "NOT_FOUND",
		6:  "ALREADY_EXISTS",
		7:  "PERMISSION_DENIED",
		8:  "UNAUTHENTICATED",
		9:  "RESOURCE_EXHAUSTED",
		10: "FAILED_PRECONDITION",
		11: "ABORTED",
		12: "OUT_OF_RANGE",
		13: "UNIMPLEMENTED",
		14: "INTERNAL",
		15: "UNAVAILABLE",
		16: "DATA_LOSS",
	}
	Response_Status_Code_value = map[string]int32{
		"OK":                  0,
		"CANCELLED":           1,
		"UNKNOWN":             2,
		"INVALID_ARGUMENT":    3,
		"DEADLINE_EXCEEDED":   4,
		"NOT_FOUND":           5,
		"ALREADY_EXISTS":      6,
		"PERMISSION_DENIED":   7,
		"UNAUTHENTICATED":     8,
		"RESOURCE_EXHAUSTED":  9,
		"FAILED_PRECONDITION": 10,
		"ABORTED":             11,
		"OUT_OF_RANGE":        12,
		"UNIMPLEMENTED":       13,
		"INTERNAL":            14,
		"UNAVAILABLE":         15,
		"DATA_LOSS":           16,
	}
)

func (x Response_Status_Code) Enum() *Response_Status_Code {
	p := new(Response_Status_Code)
	*p = x
	return p
}

func (x Response_Status_Code) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Response_Status_Code) Descriptor() protoreflect.EnumDescriptor {
	return file_udsver_proto_enumTypes[0].Descriptor()
}

func (Response_Status_Code) Type() protoreflect.EnumType {
	return &file_udsver_proto_enumTypes[0]
}

func (x Response_Status_Code) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Response_Status_Code.Descriptor instead.
func (Response_Status_Code) EnumDescriptor() ([]byte, []int) {
	return file_udsver_proto_rawDescGZIP(), []int{1, 0, 0}
}

type Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Request) Reset() {
	*x = Request{}
	mi := &file_udsver_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Request) ProtoMessage() {}

func (x *Request) ProtoReflect() protoreflect.Message {
	mi := &file_udsver_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Request.ProtoReflect.Descriptor instead.
func (*Request) Descriptor() ([]byte, []int) {
	return file_udsver_proto_rawDescGZIP(), []int{0}
}

func (x *Request) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

type Response struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Status        *Response_Status       `protobuf:"bytes,1,opt,name=status,proto3" json:"status,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Response) Reset() {
	*x = Response{}
	mi := &file_udsver_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Response) ProtoMessage() {}

func (x *Response) ProtoReflect() protoreflect.Message {
	mi := &file_udsver_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Response.ProtoReflect.Descriptor instead.
func (*Response) Descriptor() ([]byte, []int) {
	return file_udsver_proto_rawDescGZIP(), []int{1}
}

func (x *Response) GetStatus() *Response_Status {
	if x != nil {
		return x.Status
	}
	return nil
}

// TODO Maybe replace with googleapis
type Response_Status struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Code          Response_Status_Code   `protobuf:"varint,1,opt,name=code,proto3,enum=udsver.v1.Response_Status_Code" json:"code,omitempty"`
	Message       string                 `protobuf:"bytes,2,opt,name=message,proto3" json:"message,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Response_Status) Reset() {
	*x = Response_Status{}
	mi := &file_udsver_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Response_Status) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Response_Status) ProtoMessage() {}

func (x *Response_Status) ProtoReflect() protoreflect.Message {
	mi := &file_udsver_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Response_Status.ProtoReflect.Descriptor instead.
func (*Response_Status) Descriptor() ([]byte, []int) {
	return file_udsver_proto_rawDescGZIP(), []int{1, 0}
}

func (x *Response_Status) GetCode() Response_Status_Code {
	if x != nil {
		return x.Code
	}
	return Response_Status_OK
}

func (x *Response_Status) GetMessage() string {
	if x != nil {
		return x.Message
	}
	return ""
}

var File_udsver_proto protoreflect.FileDescriptor

const file_udsver_proto_rawDesc = "" +
	"\n" +
	"\fudsver.proto\x12\tudsver.v1\"\x1d\n" +
	"\aRequest\x12\x12\n" +
	"\x04name\x18\x01 \x01(\tR\x04name\"\xd2\x03\n" +
	"\bResponse\x122\n" +
	"\x06status\x18\x01 \x01(\v2\x1a.udsver.v1.Response.StatusR\x06status\x1a\x91\x03\n" +
	"\x06Status\x123\n" +
	"\x04code\x18\x01 \x01(\x0e2\x1f.udsver.v1.Response.Status.CodeR\x04code\x12\x18\n" +
	"\amessage\x18\x02 \x01(\tR\amessage\"\xb7\x02\n" +
	"\x04Code\x12\x06\n" +
	"\x02OK\x10\x00\x12\r\n" +
	"\tCANCELLED\x10\x01\x12\v\n" +
	"\aUNKNOWN\x10\x02\x12\x14\n" +
	"\x10INVALID_ARGUMENT\x10\x03\x12\x15\n" +
	"\x11DEADLINE_EXCEEDED\x10\x04\x12\r\n" +
	"\tNOT_FOUND\x10\x05\x12\x12\n" +
	"\x0eALREADY_EXISTS\x10\x06\x12\x15\n" +
	"\x11PERMISSION_DENIED\x10\a\x12\x13\n" +
	"\x0fUNAUTHENTICATED\x10\b\x12\x16\n" +
	"\x12RESOURCE_EXHAUSTED\x10\t\x12\x17\n" +
	"\x13FAILED_PRECONDITION\x10\n" +
	"\x12\v\n" +
	"\aABORTED\x10\v\x12\x10\n" +
	"\fOUT_OF_RANGE\x10\f\x12\x11\n" +
	"\rUNIMPLEMENTED\x10\r\x12\f\n" +
	"\bINTERNAL\x10\x0e\x12\x0f\n" +
	"\vUNAVAILABLE\x10\x0f\x12\r\n" +
	"\tDATA_LOSS\x10\x102:\n" +
	"\x06Verify\x120\n" +
	"\x05Check\x12\x12.udsver.v1.Request\x1a\x13.udsver.v1.ResponseB\tZ\a./protob\x06proto3"

var (
	file_udsver_proto_rawDescOnce sync.Once
	file_udsver_proto_rawDescData []byte
)

func file_udsver_proto_rawDescGZIP() []byte {
	file_udsver_proto_rawDescOnce.Do(func() {
		file_udsver_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_udsver_proto_rawDesc), len(file_udsver_proto_rawDesc)))
	})
	return file_udsver_proto_rawDescData
}

var file_udsver_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_udsver_proto_msgTypes = make([]protoimpl.MessageInfo, 3)
var file_udsver_proto_goTypes = []any{
	(Response_Status_Code)(0), // 0: udsver.v1.Response.Status.Code
	(*Request)(nil),           // 1: udsver.v1.Request
	(*Response)(nil),          // 2: udsver.v1.Response
	(*Response_Status)(nil),   // 3: udsver.v1.Response.Status
}
var file_udsver_proto_depIdxs = []int32{
	3, // 0: udsver.v1.Response.status:type_name -> udsver.v1.Response.Status
	0, // 1: udsver.v1.Response.Status.code:type_name -> udsver.v1.Response.Status.Code
	1, // 2: udsver.v1.Verify.Check:input_type -> udsver.v1.Request
	2, // 3: udsver.v1.Verify.Check:output_type -> udsver.v1.Response
	3, // [3:4] is the sub-list for method output_type
	2, // [2:3] is the sub-list for method input_type
	2, // [2:2] is the sub-list for extension type_name
	2, // [2:2] is the sub-list for extension extendee
	0, // [0:2] is the sub-list for field type_name
}

func init() { file_udsver_proto_init() }
func file_udsver_proto_init() {
	if File_udsver_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_udsver_proto_rawDesc), len(file_udsver_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   3,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_udsver_proto_goTypes,
		DependencyIndexes: file_udsver_proto_depIdxs,
		EnumInfos:         file_udsver_proto_enumTypes,
		MessageInfos:      file_udsver_proto_msgTypes,
	}.Build()
	File_udsver_proto = out.File
	file_udsver_proto_goTypes = nil
	file_udsver_proto_depIdxs = nil
}
