// Copyright (c) 2025 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package routetable

// IfacesToARPLen returns the size of the ifacesToARP set.
func (r *RouteTable) IfacesToARPLen() int {
	return r.ifacesToARP.Len()
}

// IfacesToARPAdd adds an interface to the ifacesToARP set, simulating
// the effect of a resync or interface state change.
func (r *RouteTable) IfacesToARPAdd(ifaceName string) {
	r.ifacesToARP.Add(ifaceName)
}

// GraceInfoCount returns the number of interface grace-period records
// currently being tracked.
func (r *RouteTable) GraceInfoCount() int {
	return len(r.ifaceIndexToGraceInfo)
}
