// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by lister-gen. DO NOT EDIT.

package v3

import (
	projectcalicov3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	labels "k8s.io/apimachinery/pkg/labels"
	listers "k8s.io/client-go/listers"
	cache "k8s.io/client-go/tools/cache"
)

// TierLister helps list Tiers.
// All objects returned here must be treated as read-only.
type TierLister interface {
	// List lists all Tiers in the indexer.
	// Objects returned here must be treated as read-only.
	List(selector labels.Selector) (ret []*projectcalicov3.Tier, err error)
	// Get retrieves the Tier from the index for a given name.
	// Objects returned here must be treated as read-only.
	Get(name string) (*projectcalicov3.Tier, error)
	TierListerExpansion
}

// tierLister implements the TierLister interface.
type tierLister struct {
	listers.ResourceIndexer[*projectcalicov3.Tier]
}

// NewTierLister returns a new TierLister.
func NewTierLister(indexer cache.Indexer) TierLister {
	return &tierLister{listers.New[*projectcalicov3.Tier](indexer, projectcalicov3.Resource("tier"))}
}
