/**
 * Foregrounds for text and icons
 *
 * Reference: https://designsystem.backbase.com/latest/design-tokens/semantic-colors/foreground-colors-jo60gSXd
 */

export default {
    'experimental-token-fg-default': {
        _light: 'experimental-color-neutral.1200',
        _dark: 'experimental-color-neutral.50',
    },
    'experimental-token-fg-inverted': {
        _light: 'experimental-color-neutral.0',
        _dark: 'experimental-color-neutral.1200',
    },
    'experimental-token-fg-support': {
        _light: 'experimental-color-neutral.800',
        _dark: 'experimental-color-neutral.200',
    },
    'experimental-token-fg-subtle': {
        _light: 'experimental-color-neutral.600',
        _dark: 'experimental-color-neutral.300',
    },
    'experimental-token-fg-disabled': {
        _light: 'experimental-color-neutral.400',
        _dark: 'experimental-color-neutral.400',
    },
    'experimental-token-fg-brand': {
        _light: 'experimental-color-blue.900',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-fg-brand:hovered': {
        _light: 'experimental-color-blue.800',
        _dark: 'experimental-color-medium-gold.300',
    },
    'experimental-token-fg-brand:pressed': {
        _light: 'experimental-color-blue.700',
        _dark: 'experimental-color-medium-gold.200',
    },
    'experimental-token-fg-warning': {
        _light: 'experimental-color-medium-gold.500',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-fg-info': {
        _light: 'experimental-color-medium-blue.600',
        _dark: 'experimental-color-medium-blue.300',
    },
    'experimental-token-fg-danger': {
        _light: 'experimental-color-red.600',
        _dark: 'experimental-color-red.400',
    },
    'experimental-token-fg-success': {
        _light: 'experimental-color-green.600',
        _dark: 'experimental-color-green.400',
    },
};
