// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by client-gen. DO NOT EDIT.

package fake

import (
	v3 "github.com/projectcalico/api/pkg/apis/projectcalico/v3"
	projectcalicov3 "github.com/projectcalico/api/pkg/client/applyconfiguration_generated/projectcalico/v3"
	typedprojectcalicov3 "github.com/projectcalico/api/pkg/client/clientset_generated/clientset/typed/projectcalico/v3"
	gentype "k8s.io/client-go/gentype"
)

// fakeStagedNetworkPolicies implements StagedNetworkPolicyInterface
type fakeStagedNetworkPolicies struct {
	*gentype.FakeClientWithListAndApply[*v3.StagedNetworkPolicy, *v3.StagedNetworkPolicyList, *projectcalicov3.StagedNetworkPolicyApplyConfiguration]
	Fake *FakeProjectcalicoV3
}

func newFakeStagedNetworkPolicies(fake *FakeProjectcalicoV3, namespace string) typedprojectcalicov3.StagedNetworkPolicyInterface {
	return &fakeStagedNetworkPolicies{
		gentype.NewFakeClientWithListAndApply[*v3.StagedNetworkPolicy, *v3.StagedNetworkPolicyList, *projectcalicov3.StagedNetworkPolicyApplyConfiguration](
			fake.Fake,
			namespace,
			v3.SchemeGroupVersion.WithResource("stagednetworkpolicies"),
			v3.SchemeGroupVersion.WithKind("StagedNetworkPolicy"),
			func() *v3.StagedNetworkPolicy { return &v3.StagedNetworkPolicy{} },
			func() *v3.StagedNetworkPolicyList { return &v3.StagedNetworkPolicyList{} },
			func(dst, src *v3.StagedNetworkPolicyList) { dst.ListMeta = src.ListMeta },
			func(list *v3.StagedNetworkPolicyList) []*v3.StagedNetworkPolicy {
				return gentype.ToPointerSlice(list.Items)
			},
			func(list *v3.StagedNetworkPolicyList, items []*v3.StagedNetworkPolicy) {
				list.Items = gentype.FromPointerSlice(items)
			},
		),
		fake,
	}
}
