export const linkStyles = {
    color: 'tigeraBlueMedium40',
    svg: {
        fill: 'tigeraBlueMedium40',
    },
    textDecoration: 'underline',
    fontSize: 'sm',
};

export const alertDescriptionStyles = {
    ml: 'auto',
    fontSize: 'sm',
    fontWeight: 'medium',
};
