/**
 * Border colours
 *
 * Reference: https://designsystem.backbase.com/latest/design-tokens/semantic-colors/border-colors-ycZQpqDi
 */

export default {
    'experimental-token-border-default': {
        _light: 'experimental-color-neutral.100',
        _dark: 'experimental-color-neutral.800',
    },
    'experimental-token-border-default:hovered': {
        _light: 'experimental-color-neutral.200',
        _dark: 'experimental-color-neutral.700',
    },
    'experimental-token-border-bold': {
        _light: 'experimental-color-neutral.200',
        _dark: 'experimental-color-neutral.500',
    },
    'experimental-token-border-selected': {
        _light: 'experimental-color-blue.900',
        _dark: 'experimental-color-medium-gold.300',
    },
    'experimental-token-border-brand': {
        _light: 'experimental-color-blue.900',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-border-brand:hovered': {
        _light: 'experimental-color-blue.900',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-border-brand:pressed': {
        _light: 'experimental-color-blue.700',
        _dark: 'experimental-color-medium-gold.200',
    },
    'experimental-token-border-warning': {
        _light: 'experimental-color-medium-gold.500',
        _dark: 'experimental-color-medium-gold.400',
    },
    'experimental-token-border-info': {
        _light: 'experimental-color-medium-blue.600',
        _dark: 'experimental-color-medium-blue.300',
    },
    'experimental-token-border-danger': {
        _light: 'experimental-color-red.600',
        _dark: 'experimental-color-red.400',
    },
    'experimental-token-border-success': {
        _light: 'experimental-color-green.600',
        _dark: 'experimental-color-green.400',
    },
};
