// Copyright (c) 2026 Tigera, Inc. All rights reserved.

// Code generated by applyconfiguration-gen. DO NOT EDIT.

package v3

// KubeControllersConfigurationStatusApplyConfiguration represents a declarative configuration of the KubeControllersConfigurationStatus type for use
// with apply.
//
// KubeControllersConfigurationStatus represents the status of the configuration. It's useful for admins to
// be able to see the actual config that was applied, which can be modified by environment variables on the
// kube-controllers process.
type KubeControllersConfigurationStatusApplyConfiguration struct {
	// RunningConfig contains the effective config that is running in the kube-controllers pod, after
	// merging the API resource with any environment variables.
	RunningConfig *KubeControllersConfigurationSpecApplyConfiguration `json:"runningConfig,omitempty"`
	// EnvironmentVars contains the environment variables on the kube-controllers that influenced
	// the RunningConfig.
	EnvironmentVars map[string]string `json:"environmentVars,omitempty"`
}

// KubeControllersConfigurationStatusApplyConfiguration constructs a declarative configuration of the KubeControllersConfigurationStatus type for use with
// apply.
func KubeControllersConfigurationStatus() *KubeControllersConfigurationStatusApplyConfiguration {
	return &KubeControllersConfigurationStatusApplyConfiguration{}
}

// WithRunningConfig sets the RunningConfig field in the declarative configuration to the given value
// and returns the receiver, so that objects can be built by chaining "With" function invocations.
// If called multiple times, the RunningConfig field is set to the value of the last call.
func (b *KubeControllersConfigurationStatusApplyConfiguration) WithRunningConfig(value *KubeControllersConfigurationSpecApplyConfiguration) *KubeControllersConfigurationStatusApplyConfiguration {
	b.RunningConfig = value
	return b
}

// WithEnvironmentVars puts the entries into the EnvironmentVars field in the declarative configuration
// and returns the receiver, so that objects can be build by chaining "With" function invocations.
// If called multiple times, the entries provided by each call will be put on the EnvironmentVars field,
// overwriting an existing map entries in EnvironmentVars field with the same key.
func (b *KubeControllersConfigurationStatusApplyConfiguration) WithEnvironmentVars(entries map[string]string) *KubeControllersConfigurationStatusApplyConfiguration {
	if b.EnvironmentVars == nil && len(entries) > 0 {
		b.EnvironmentVars = make(map[string]string, len(entries))
	}
	for k, v := range entries {
		b.EnvironmentVars[k] = v
	}
	return b
}
