export const gridStyles = {
    w: '100%',
    h: '100vh',
    minWidth: '1024px',
    overflowY: 'hidden',
    bg: 'tigera-color-surface-secondary',
    gridTemplateAreas: '"promo-banner" "header" "main"',
    gridTemplateRows: 'max-content minmax(54px, max-content) auto',
};
