// Copyright (c) 2025 Tigera, Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Package header contains constants and utilities related to http headers. Any header or header value that's used within
// this module must reference the definitions in here.
package header

const (
	ContentType              = "Content-Type"
	ApplicationJSON          = "application/json; charset=utf-8"
	XRequestId               = "X-Request-Id"
	CacheControl             = "Cache-Control"
	Connection               = "Connection"
	AccessControlAllowOrigin = "Access-Control-Allow-Origin"
	NoCache                  = "no-cache"
	KeepAlive                = "keep-alive"
	TextEventStream          = "text/event-stream"
)
